/*
 * @package secureLogin
 * @author Sebastian Tschan
 * @copyright (c) Sebastian Tschan
 * @license GNU General Public License
 * @link https://blueimp.net/mozilla/
 */

var secureLogin = {
	
	// Helper var to recognize the Secure Login object:
	secureLoginObject: true,
	// Secure Logins preferences branch:
	secureLoginPrefs: null,
	// The progress listener:
	progressListener: null,
	// Variable to define if the progress listener has been registered to the browser:
	isProgressListenerRegistered: null,
	// Helper var to remember original autofillForms setting (this has nothing to to with the extension autofillForms@blueimp.net:
	autofillForms: null,
	// Valid logins list:
	secureLogins: null,
	// Helper list to store the form index:
	secureLoginsFormIndex: null,
	// Helper list to store the document window (frame):
	secureLoginsWindow: null,
	// Helper list to store the username field:
	secureLoginsUserField: null,
	// Helper list to store the password field:
	secureLoginsPassField: null,
	// Defines if form index is to be shown in selection prompt:
	showFormIndex: null,
	// Object containing the shortcut information (modifiers, key or keycode):
	shortcut: null,
	// Helper object for storing a new shortcut before applying it to the shortcut object:
	shortcutNew: null,
	// Helper var to remember a failed bookmark-login attempt:
	failedBookmarkLogin: null,
	// Event listener for the content area context menu:
	contentAreaContextMenuEventListener: null,
	// Temporary exceptions list copy for the exceptions window:
	exceptions: null,
	// The exceptions tree object:
	exceptionsTree: null,
	// The exceptions treeView object
	exceptionsTreeView: null,
	// The exceptions treeBox object:
	exceptionsTreeBox: null,
	// The exceptions treeSelection object:
	exceptionsTreeSelection: null,
	// Determines if exceptions sort is to be ascending or descending:
	exceptionsAscending: null,

	init: function() {
		
		// Add a preferences observer to the secureLogin preferences branch:
		this.getSecureLoginPrefs().QueryInterface(Components.interfaces.nsIPrefBranch2);
     	this.getSecureLoginPrefs().addObserver('', this, false);
		
		// Implement the listener methods:
		this.progressListener = {
			QueryInterface: function(aIID) {
				if(aIID.equals(Components.interfaces.nsIWebProgressListener) ||
					aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
					aIID.equals(Components.interfaces.nsISupports))
					return this;
				throw Components.results.NS_NOINTERFACE;
			},
			onStateChange:function(aProgress,aRequest,aFlag,aStatus) {
				// Update status when load finishes:
				if(aFlag & Components.interfaces.nsIWebProgressListener.STATE_STOP) {
					this.parent.updateStatus(aProgress,aRequest,null,aFlag,aStatus);
				}
			},
			// Update status when location changes (tab change):
			onLocationChange:function(aProgress,aRequest,aLocation){
				this.parent.updateStatus(aProgress,aRequest,aLocation,null,null);
			},
			onProgressChange:function(a,b,c,d,e,f){},
			onStatusChange:function(a,b,c,d){},
			onSecurityChange:function(a,b,c){},
			onLinkIconAvailable:function(a){}
		}
		// Set the secureLogin object as parent:
		this.progressListener.parent = this;

		// Implement the event listener for the content area context menu:
		this.contentAreaContextMenuEventListener = function() {
			secureLogin.initContentAreaContextMenu();
		}
		
		// Initialize the preferences settings:
		this.initializePrefs();
					
		// Remove the "Remember passwords exceptions" menuitem if firefox version is not 2.0 or later:
		if(this.getVersionComparator().compare(this.getAppInfo().version, '2.0') < 0) {
			if(document.getElementById('passexceptionsContextItem'))
				document.getElementById('passexceptionsContextItem').parentNode
					.removeChild(document.getElementById('passexceptionsContextItem'));
			if(document.getElementById('passexceptionsToolsMenuItem'))
				document.getElementById('passexceptionsToolsMenuItem').parentNode
					.removeChild(document.getElementById('passexceptionsToolsMenuItem'));
		}
  	},

	initContentAreaContextMenu: function() {
		var cm0 = document.getElementById('secureLoginContextMenuItem');
		var cm1 = document.getElementById('secureLoginContextMenuMenu');
		var cm2 = document.getElementById('secureLoginContextMenuSeparator1');
		var cm3 = document.getElementById('secureLoginContextMenuSeparator2');
		if(cm0 && gContextMenu) {
			if(this.getSecureLoginPrefs().getBoolPref('hideContextMenuItem')
				|| gContextMenu.isContentSelected
				|| gContextMenu.onTextInput
				|| gContextMenu.onImage
				|| gContextMenu.onLink
				|| gContextMenu.onCanvas
				|| gContextMenu.onMathML
				|| !this.getDoc().forms
				|| !this.getDoc().forms.length) {
				cm0.hidden = true;
				cm1.hidden = true;
				cm2.hidden = true;
				cm3.hidden = true;
			} else {
				// Search for valid logins and outline login fields if not done automatically:
				if(!this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {
					this.searchLoginsInitialize();
				}
				if(!this.secureLogins || !this.secureLogins.length) {
					cm0.hidden = true;
					cm1.hidden = true;
					cm2.hidden = true;
					cm3.hidden = true;
				} else {
					// Determine if no master password is set or the user has already been authenticated:
					var masterPasswordRequired = true;
					if(!this.getMasterSecurityDevice().getInternalKeyToken().needsLogin()
						|| this.getMasterSecurityDevice().getInternalKeyToken().isLoggedIn()) {
						masterPasswordRequired = false;
					}
					// Show the menu or the menu item depending on the numer of logins and the MSD status:
					if(this.secureLogins.length > 1 && !masterPasswordRequired) {
						cm0.hidden = true;
						cm1.hidden = false;
					} else {
						cm0.hidden = false;
						cm1.hidden = true;
					}
					// Show menuseparators if not already separated:
					if(this.isPreviousNodeSeparated(cm2)) {
						cm2.hidden = true;
					} else {
						cm2.hidden = false;
					}
					if(this.isNextNodeSeparated(cm3)) {
						cm3.hidden = true;
					} else {
						cm3.hidden = false;
					}
				}
			}
		}
	},
	
	isNextNodeSeparated: function(node) {
		while(node) {
			node = node.nextSibling
			if(node.hidden) {
				continue;
			}
			if(node.nodeName == 'menuseparator') {
				return true;
			} else {
				return false;
			}
		}
		return true;
	},
	
	isPreviousNodeSeparated: function(node) {
		while(node) {
			node = node.previousSibling;
			if(node.hidden) {
				continue;
			}
			if(node.nodeName == 'menuseparator') {
				return true;
			} else {
				return false;
			}
		}
		return true;
	},
  	
	initializePrefs: function() {
	
		// Disable the prefilling of login forms if enabled, remember status:
		try {
			if(this.getVersionComparator().compare(this.getAppInfo().version, '2.*') < 0) {
				// Firefox version 1.5 - 2.0.0.*:
				if(this.getPrefManager().getBranch('').getBoolPref('signon.prefillForms')) {
					this.getPrefManager().getBranch('').setBoolPref('signon.prefillForms', false);
					this.autofillForms = true;
				} else {
					this.autofillForms = false;
				}
			} else {
				// Firefox version 3 and prereleases:
				if(this.getPrefManager().getBranch('').getBoolPref('signon.autofillForms')) {
					this.getPrefManager().getBranch('').setBoolPref('signon.autofillForms', false);
					this.autofillForms = true;
				} else {
					this.autofillForms = false;
				}
			}
		} catch(e) {
			//alert(e);
		}

		// Add the progress listener to the browser, set the Secure Login icons:
		this.searchLoginsOnloadUpdate();

		// Set the keyboard shortcut:
		this.updateShortcut();
	
		// Initialize toolbar and statusbar icons and tools and context menus:
		this.hideToolbarButtonUpdate();
		this.hideToolbarButtonMenuUpdate();
		this.hideStatusbarIconUpdate();
		this.hideToolsMenuUpdate();
		this.hideContextMenuItemUpdate();
		this.javascriptProtectionUpdate();
	},

	observe: function(subject, topic, data) {
		// Only observe preferences changes:
		if (topic != 'nsPref:changed')
			return;
		switch(data) {
			case 'shortcut':
				this.updateShortcut();
				break;
			case 'hideContextMenuItem':
				this.hideContextMenuItemUpdate();
				break;
			case 'hideToolsMenu':
				this.hideToolsMenuUpdate();
				break;
			case 'hideStatusbarIcon':
				this.hideStatusbarIconUpdate();
				break;
			case 'hideToolbarButton':
				this.hideToolbarButtonUpdate();
				break;
			case 'hideToolbarButtonMenu':
				this.hideToolbarButtonMenuUpdate();
				break;
			case 'searchLoginsOnload':
				this.searchLoginsOnloadUpdate();
				break;
			case 'secureLoginBookmarks':
				this.secureLoginBookmarksUpdate();
				break;
			case 'highlightColor':
				this.highlightColorUpdate();
				break;
			case 'javascriptProtection':
				this.javascriptProtectionUpdate();
				break;
		}
	},
	
	changePref: function(doc, event, pref) {
		// Attribute 'checked' is empty or true, setting must be false or true:
		this.getSecureLoginPrefs().setBoolPref(
			pref,
			!!event.target.getAttribute('checked')
		);
	},

	progressListenerUpdate: function() {
		if(!this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload') && !this.getSecureLoginPrefs().getBoolPref('secureLoginBookmarks')) {
			// Remove the listener from the browser object (if added previously):
			try {
				this.getBrowser().removeProgressListener(this.progressListener);
				this.isProgressListenerRegistered = false;
			} catch(e) {}
		} else if(!this.isProgressListenerRegistered && 
			(this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload') || this.getSecureLoginPrefs().getBoolPref('secureLoginBookmarks'))) {
			// Add the progress listener to the browser object (if not added previously):
			try {
				this.getBrowser().addProgressListener(
					this.progressListener,
					Components.interfaces.nsIWebProgress.NOTIFY_LOCATION | Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT
				);
				this.isProgressListenerRegistered = true;
			} catch(e) {}
		}
	},
		
	searchLoginsOnloadUpdate: function() {
		this.progressListenerUpdate();

		if(this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {		
			// Search for valid logins and outline login fields:
			this.searchLoginsInitialize();
		} else {
			// Always highlight the Secure Login icons, when not searching for valid logins automatically:
			if(document.getElementById('secureLoginPanelIcon')) {
				document.getElementById('secureLoginPanelIcon').setAttribute('class', 'statusbarpanel-menu-iconic secureLoginIcon');
			}
			if(document.getElementById('secureLoginButton')) {
				document.getElementById('secureLoginButton').setAttribute('class', 'toolbarbutton-1');
			}		
		}
	},

	secureLoginBookmarksUpdate: function() {
		if(this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash') == '#secureLoginBookmark') {
			// Create a random Secure Login Bookmark hash (anchor) if the default is still set:
			// This slightly increases security and avoids unwanted auto-logins
			this.getSecureLoginPrefs().setCharPref('secureLoginBookmarkHash', '#slb'+Math.ceil(Math.random()*1000000000));
		}
		
		this.progressListenerUpdate();
	},

	highlightColorUpdate: function() {
		if(this.secureLoginsPassField) {
			// The outline style:
			var outlineStyle = ''
								+ this.getSecureLoginPrefs().getIntPref('highlightOutlineWidth')
								+ 'px '
								+ this.getSecureLoginPrefs().getCharPref('highlightOutlineStyle')
								+ ' '
								+ this.getSecureLoginPrefs().getCharPref('highlightColor');
			
			// Update the outlined form fields:
			for(var i=0; i<this.secureLoginsPassField.length; i++) {
				// Outline the username field if existing:
				if(this.secureLoginsUserField[i])
					this.secureLoginsUserField[i].style.outline = outlineStyle;
				// Outline the password field if existing:
				if(this.secureLoginsPassField[i])
					this.secureLoginsPassField[i].style.outline = outlineStyle;
			}
		}
	},

	installToolbarButton: function(buttonID, beforeNodeID, toolbarID) {
		beforeNodeID = beforeNodeID ? beforeNodeID : 'urlbar-container';
		toolbarID = toolbarID ? toolbarID : 'navigation-toolbar';
		if(!document.getElementById(buttonID)) {
			var toolbar = document.getElementById(toolbarID);
			if(!toolbar) {
				// Firefox < 3:
				toolbar = document.getElementById('nav-bar');
			}			
			if(toolbar && 'insertItem' in toolbar) {
				var beforeNode = document.getElementById(beforeNodeID);
				if(beforeNode && beforeNode.parentNode != toolbar) {
					beforeNode = null;
				}
				// Insert before the given node or at the end of the toolbar if the node is not available:
				toolbar.insertItem(buttonID, beforeNode, null, false);
				toolbar.setAttribute('currentset', toolbar.currentSet);
				document.persist(toolbar.id, 'currentset');
			}
		}
	},
	
	hideToolbarButtonUpdate: function() {
		if(!document.getElementById('secureLoginButton')) {
			// Add the toolbar button to the toolbar if it is not already present:
			this.installToolbarButton('secureLoginButton');
		}
		if(document.getElementById('secureLoginButton')) {
			document.getElementById('secureLoginButton').setAttribute(
				'hidden',
				this.getSecureLoginPrefs().getBoolPref('hideToolbarButton')
			);
		}
	},

	hideToolbarButtonMenuUpdate: function() {
		if(document.getElementById('secureLoginButton')) {
			if(this.getSecureLoginPrefs().getBoolPref('hideToolbarButtonMenu')) {
				document.getElementById('secureLoginButton').removeAttribute('type');
			} else {
				document.getElementById('secureLoginButton').setAttribute('type','menu-button');
			}
		}
	},
	
	hideStatusbarIconUpdate: function() {
		// Change the statusbar icon visibility:
		if(document.getElementById('secureLoginPanelIcon')) {
			document.getElementById('secureLoginPanelIcon').setAttribute(
				'hidden',
				this.getSecureLoginPrefs().getBoolPref('hideStatusbarIcon')
			);
		}
	},
	
	hideToolsMenuUpdate: function() {
		// Change the tools menu visibility:
		if(document.getElementById('secureLoginToolsMenu')) {
			document.getElementById('secureLoginToolsMenu').setAttribute(
				'hidden',
				this.getSecureLoginPrefs().getBoolPref('hideToolsMenu')
			);
		}
	},
	
	hideContextMenuItemUpdate: function() {
		if(document.getElementById('contentAreaContextMenu')) {
			if(!this.getSecureLoginPrefs().getBoolPref('hideContextMenuItem')) {
				// Add the content area context menu listener:
				document.getElementById('contentAreaContextMenu').addEventListener(
					'popupshowing',
					this.contentAreaContextMenuEventListener,
					false
				);
			} else {
				// Hide the SL contentare context menu entries and remove the content area context menu listener:
				var cm0 = document.getElementById('secureLoginContextMenuItem');
				var cm1 = document.getElementById('secureLoginContextMenuMenu');
				var cm2 = document.getElementById('secureLoginContextMenuSeparator1');
				var cm3 = document.getElementById('secureLoginContextMenuSeparator2');
				if(cm0) {
					cm0.hidden = true;
					cm1.hidden = true;
					cm2.hidden = true;
					cm3.hidden = true;
				}
				document.getElementById('contentAreaContextMenu').removeEventListener(
					'popupshowing',
					this.contentAreaContextMenuEventListener,
					false
				);
			}
		}
	},
	
	javascriptProtectionUpdate: function() {
		// Update the statusbar icon context menu JavaScript protection item:
		if(document.getElementById('javascriptProtectionContextItem')) {
			document.getElementById('javascriptProtectionContextItem').setAttribute(
				'checked',
				this.getSecureLoginPrefs().getBoolPref('javascriptProtection')
			);
		}
		// Update the tools menu JavaScript protection item:
		if(document.getElementById('javascriptProtectionToolsMenuItem')) {
			document.getElementById('javascriptProtectionToolsMenuItem').setAttribute(
				'checked',
				this.getSecureLoginPrefs().getBoolPref('javascriptProtection')
			);
		}
		// Update the tools menu JavaScript protection item:
		if(document.getElementById('javascriptProtectionButtonMenuItem')) {
			document.getElementById('javascriptProtectionButtonMenuItem').setAttribute(
				'checked',
				this.getSecureLoginPrefs().getBoolPref('javascriptProtection')
			);
		}
	},

	updateStatus: function(aProgress,aRequest,aLocation,aFlag,aStatus) {		
		if(this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {
			// Initialize the recursive search for logins on the current window:
			this.searchLoginsInitialize(aProgress.DOMWindow);
			
			if(this.getSecureLoginPrefs().getBoolPref('autoLogin')
				&& this.secureLogins && this.secureLogins.length > 0) {
				// Auto-Login if enabled and logins have been found:
				this.login(aProgress.DOMWindow);
			}
		}

		if(this.getSecureLoginPrefs().getBoolPref('secureLoginBookmarks') 
			&& !this.getSecureLoginPrefs().getBoolPref('autoLogin')) {
			// Auto-Login if the current URL is a Secure Login Bookmark:
			this.bookmarkLogin(aProgress.DOMWindow);
		}
	},
	
	bookmarkLogin: function(win) {
		// Check for first four characters of Secure Login anchor (hash):
		if(this.getDoc(win) && this.getDoc(win).location && this.getDoc(win).location.hash
			&& this.getDoc(win).location.hash.substr(0, 4) == this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash').substr(0, 4)) {
			
			// Check for complete Secure Login anchor (hash):
			var index = this.getDoc(win).location.hash.indexOf(this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash'));
			if(index == 0) {				
				var bookmarkLoginIndex = parseInt(
					this.getDoc(win).location.hash.substr(
						this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash').length
					)
				);
				if(!isNaN(bookmarkLoginIndex)) {
					// Auto-Login using the bookmarkLoginIndex:
					this.login(win, bookmarkLoginIndex);
				} else {
					// Auto-Login:
					this.login(win);
				}
			} else {
				// Remember failed bookmark-login attempt:
				this.failedBookmarkLogin = true;
			}
		}
	},
	
	searchLoginsInitialize: function(win) {
		if(!win || !win.document) {
			win = this.getWin();
		}
		
		// New valid logins list:
		this.secureLogins = new Array();
		// New helper list to store the form index:
		this.secureLoginsFormIndex = new Array();
		// New helper list to store the document window (frame):
		this.secureLoginsWindow = new Array();
		// New helper list to store the username field:
		this.secureLoginsUserField = new Array();
		// New helper list to store the password field:
		this.secureLoginsPassField = new Array();
		
		// Show form index only if more than one valid login form is found:
		this.showFormIndex = false;
		
		// Search for valid logins on the given window:
		this.searchLogins(win);


		if(this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {
			// Update the icons:
			if(this.secureLogins.length > 0) {
	
				if(document.getElementById('secureLoginPanelIcon')) {
					document.getElementById('secureLoginPanelIcon').setAttribute('class', 'statusbarpanel-menu-iconic secureLoginIcon');
				}
				if(document.getElementById('secureLoginButton')) {
					document.getElementById('secureLoginButton').setAttribute('class', 'toolbarbutton-1');
				}
				
		     	// Play sound notification:
		     	if(this.getSecureLoginPrefs().getBoolPref('playLoginFoundSound')) {
		     		this.playSound('loginFoundSoundFileName');
		     	}
		     		
			} else {
				// No logins found:
				if(document.getElementById('secureLoginPanelIcon')) {
					document.getElementById('secureLoginPanelIcon').setAttribute('class', 'statusbarpanel-menu-iconic secureLoginIconDisabled');
				}
				if(document.getElementById('secureLoginButton')) {
					document.getElementById('secureLoginButton').setAttribute('class', 'toolbarbutton-1 disabled');
				}
			}
		}
	},
	
	searchLogins: function(win) {		
		// Recursive call for all subframes:
		for(var f=0; f < win.frames.length; f++) {
			this.searchLogins(win.frames[f]);
		}
	
		// Check the location (SecureLogin is only useful for HTTP or HTTPS logins) and for existing forms:
		if(this.getDoc(win) && this.getDoc(win).location
			&& (this.getDoc(win).location.protocol == 'http:' || this.getDoc(win).location.protocol == 'https:')
			&& this.getDoc(win).forms && this.getDoc(win).forms.length > 0) {
			
			// document (current) host:
			var host = this.getDoc(win).location.protocol + '//' + this.getDoc(win).location.host;
			
			if(this.getLoginManager()) {
				// Firefox 3:
				
				var formActions = new Array();
				
				// Go through the forms:
				for(var i = 0; i < this.getDoc(win).forms.length; i++) {

					if(this.getSecureLoginPrefs().getBoolPref('skipDuplicateActionForms')) {
						// Skip this form if the same action target has already been added:
						if(this.inArray(formActions, this.getDoc(win).forms[i].action)) {
							continue;
						}
					}

					// Get the targetHost from the form "action" attribute:
					var targetHost;
					if(this.getDoc(win).forms[i].action) {
						try {
							targetHost = this.makeURI(this.getDoc(win).forms[i].action, this.getDoc(win).characterSet).prePath;							
						} catch(e) {
							// The forms seems not to have a valid "acion" attribute, continue:
							continue;
						}
					} else {
						// Forms with no "action" attribute default to submitting to their origin URL:
						targetHost = host;
					}
					
					// Getting the number of existing logins with countLogins() instead of findLogins() to avoid a Master Password prompt:
					var loginsCount = this.getLoginManager().countLogins(host, targetHost, null);
					
					// Go through the logins:
					for(var j=0; j<loginsCount; j++) {
				 		
						// Get valid login fields:
				 		var loginFields = this.getLoginFields(this.getDoc(win).forms[i], null, null);
				 		
				 		if(loginFields) {			
							
							if(this.getSecureLoginPrefs().getBoolPref('skipDuplicateActionForms')) {
								// Add the forms action target to the list:
								formActions.push(this.getDoc(win).forms[i].action);
							}

				 			// Add null as login object to the logins list to avoid a Master Password prompt:
							this.addToFoundLoginsList(null, i, win, loginFields.username, loginFields.password);
							
							// highlight login fields:
							this.highlightLoginFields(loginFields.username, loginFields.password);
				 		}
			 		}
				}
				
			} else {
				// Firefox versions before Firefox 3:
				
				// Get an enumerator for the stored logins:
				var loginsEnumerator = this.getPasswordManager().enumerator;
				
				// step through the login list:
				while (loginsEnumerator.hasMoreElements()) {
				    
					// get an nsIPasswordInternal type (which inherits from nsIPassword) out of the password manager:
				    var login = loginsEnumerator.getNext().QueryInterface(Components.interfaces.nsIPasswordInternal);
				    
					// Compare login host and document (current) host:
					if(login.host == host) {

						var formActions = new Array();

					 	// Go through the forms:
					 	for(var i = 0; i < this.getDoc(win).forms.length; i++) {
							
							if(this.getSecureLoginPrefs().getBoolPref('skipDuplicateActionForms')) {
								// Skip this form if the same action target has already been added:
								if(this.inArray(formActions, this.getDoc(win).forms[i].action)) {
									continue;
								}
							}
							
					 		// Get valid login fields:
					 		var loginFields = this.getLoginFields(this.getDoc(win).forms[i], login.userFieldName, login.passwordFieldName);
					 		
					 		if(loginFields) {			

								if(this.getSecureLoginPrefs().getBoolPref('skipDuplicateActionForms')) {
									// Add the forms action target to the list:
									formActions.push(this.getDoc(win).forms[i].action);
								}

					 			// Add valid login object to the logins list:
								this.addToFoundLoginsList(login, i, win, loginFields.username, loginFields.password);
								
								// highlight login fields:
								this.highlightLoginFields(loginFields.username, loginFields.password);
					 		}
						}
					 	
					}
				}
			}
		}
	},

	getRealLoginObjects: function() {
		// Method for Firefox 3 to get the real login objects instead of the null values:
		
		var loginObjects = new Array();		
		
		// Go through the collected dummy login objects (null values):
		for(var i=0; i < this.secureLogins.length; i++) {			
			var win = this.secureLoginsWindow[i];
			var formIndex = this.secureLoginsFormIndex[i];
			
			var host = this.getDoc(win).location.protocol + '//' + this.getDoc(win).location.host;
			var targetHost;
			if(this.getDoc(win).forms[formIndex].action) {
				try {
					targetHost = this.makeURI(this.getDoc(win).forms[formIndex].action, this.getDoc(win).characterSet).prePath;							
				} catch(e) {
					// This should not happen because we did it already when calling the searchLogins method
				}
			} else {
				// Forms with no "action" attribute default to submitting to their origin URL:
				targetHost = host;
			}
			
			try {
				// This should always return some login objects, as countLogins() had not returned 0 either:
				var logins = this.getLoginManager().findLogins({}, host, targetHost, null);
				
				loginObjects = loginObjects.concat(this.getLoginManager().findLogins({}, host, targetHost, null));
				
				// Increase the counter for the number of found logins (-1 as i++ increases the counter as well):
				i = i + logins.length -1;
			} catch(e) {
				// User cancelled master password entry, so we break out of the loop:
				break;
			}
		}
		
		return loginObjects;
	},
	
	getLoginFields: function(form, loginUsernameFieldName, loginPasswordFieldName) {
	
		// The form fields for user+pass:
		var usernameField = null;
		var passwordField = null;

		// helper var to define if the login form is a password only form:
		var inputTextFound = false;

		// The form elements list:
		var elements = form.elements;
		
		// Go through the form elements:
		for(var i = 0; i < elements.length; i++) {
			// Skip disabled elements or elements without a "name":
		    if(!elements[i].name || elements[i].disabled)
		    	continue;					
			if(elements[i].type == 'text') {
				// input of type "text" found, this is no password only form:
				inputTextFound = true;
				
				// We do not get a loginUsernameFieldName from Firefox 3:
				if(!loginUsernameFieldName) {
					// Assume the first text field followed by a password field is the username field
					// Use another loop to skip non-text fields (e.g. checkboxes) between:
					for(var j = i+1; j < elements.length; j++) {
						if(elements[j].type == 'password') {
							// Following password field found so the username field might be valid:
							usernameField = elements[i];
							break;
						}
						if(elements[j].type == 'text') {
							// Another textfield found, this might be the username field, so break out of the loop:
							break;
						}
					}
				} else {
					if(elements[i].name == loginUsernameFieldName) {
						usernameField = elements[i];
					}
				}
			} else if(elements[i].type == 'password') {
				// We do not get a loginPasswordFieldName from Firefox 3:
				if(!loginPasswordFieldName) {
					// Skip registration or password change forms (two password fields):
					if(!(elements[i+1] && elements[i+1].type == 'password')) {
						passwordField = elements[i];
					}					
				} else {
					// Skip registration or password change forms (two password fields):
					if(!(elements[i+1] && elements[i+1].type == 'password') && elements[i].name == loginPasswordFieldName) {
						passwordField = elements[i];
					}
				}
				
				// We found a password field so break out of the loop:
				break;
			}	    	
		}

		if(passwordField) {
			// If this is a password only form, no input of type "text" may be found and userFieldName must be empty:
			if(!usernameField && (inputTextFound || loginUsernameFieldName))
				return null;
		    
			var loginFields = new Object();
			loginFields.username = usernameField;
			loginFields.password = passwordField;
			
			return loginFields;
		}
		return null;
	},
	
	addToFoundLoginsList: function(loginObject, formIndex, windowObject, usernameField, passwordField) {
		var loginIndex = this.secureLogins.length;
		
		// Test if there is only one valid login form:
		if(!this.showFormIndex && loginIndex > 0 && !this.inArray(this.secureLoginsFormIndex, formIndex))
			this.showFormIndex = true;
			  
		// Save the login in the valid logins list:
		this.secureLogins[loginIndex] = loginObject;
		// Save the form index in the list:
		this.secureLoginsFormIndex[loginIndex] = formIndex;
		// Save the current document window (frame) in the list:
		this.secureLoginsWindow[loginIndex] = windowObject;
		// Save the username field in the list:
		this.secureLoginsUserField[loginIndex] = usernameField;
		// Save the password field in the list:
		this.secureLoginsPassField[loginIndex] = passwordField;	
	},
	
	highlightLoginFields: function(usernameField, passwordField) {
		// Possible style declaration, overwriting outline settings:
		var highlightStyle = this.getSecureLoginPrefs().getCharPref('highlightStyle');	

		if(!highlightStyle) {
			if(!this.getSecureLoginPrefs().getIntPref('highlightOutlineWidth')) {
				// No visible style set, return:
				return;
			}
			
			// The outline style:
			var outlineStyle = ''
								+ this.getSecureLoginPrefs().getIntPref('highlightOutlineWidth')
								+ 'px '
								+ this.getSecureLoginPrefs().getCharPref('highlightOutlineStyle')
								+ ' '
								+ this.getSecureLoginPrefs().getCharPref('highlightColor');
			
			// The outline radius:
			var outlineRadius = this.getSecureLoginPrefs().getIntPref('highlightOutlineRadius');
		}
		
		// Outline usernameField:
		if(usernameField) {
	    	// Overwrite style if set:
	    	if(highlightStyle) {
	    		usernameField.setAttribute('style', highlightStyle);
	    	} else {
	    		usernameField.style.outline = outlineStyle;
	    		if(outlineRadius) {
		    		usernameField.style.setProperty(
		    			'-moz-outline-radius',
		    			outlineRadius+'px',
		    			null
		    		);
	    		}
	    	}
	    }
	    
		// Overwrite highlight style if set:
		if(highlightStyle) {
			passwordField.setAttribute('style', highlightStyle);
		} else {
			// outline the password field:
			passwordField.style.outline = outlineStyle;
    		if(outlineRadius) {
	    		passwordField.style.setProperty(
	    			'-moz-outline-radius',
	    			outlineRadius+'px',
	    			null
	    		);
    		}
		}
	},
	
	tooltip: function(event) {
	
		// Check if the tooltip exists and if it is shown above a valid node:
		if (!document.tooltipNode || !document.tooltipNode.hasAttribute('tooltip')
			|| !(document.tooltipNode.id == 'secureLoginButton' || document.tooltipNode.id == 'secureLoginPanelIcon')) {
			// Don't show any tooltip:
			event.preventDefault();
			return;
		}
			
		// Search for valid logins and outline login fields if not done automatically:
		if(!this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {
			this.searchLoginsInitialize();
		}
			
		// Get the tooltip node:
		var tooltip = document.getElementById('secureLoginTooltip');
		// Remove all children nodes:
		while(tooltip.hasChildNodes())
			tooltip.removeChild(tooltip.firstChild);
		
		var label = document.createElement('label');
		var hbox, descr, spacer;
		
		if(this.secureLogins && this.secureLogins.length > 0) {
		
			// List of unique action urls:
			var urls = new Array();
			// Helper list to count the number of identical urls:
			var urlsCount = new Array();
			
			// Go through the forms and find the unique action urls:
			var win;
			var formIndex;
			var url;
			var foundInList;
			for(var i = 0; i < this.secureLogins.length; i++) {
				win = this.secureLoginsWindow[i];
				formIndex = this.secureLoginsFormIndex[i];
				url = this.getDoc(win).forms[formIndex].action;
				// If the url is empty, take it from the current document:
				if(!url)
					url = this.getDoc(win).baseURI;
				foundInList = false;
				// Check if the form action url is already in the list:
				for(var j = 0; j < urls.length; j++) {
					if(urls[j] == url) {
						// url already in the list, increase the counter:
						foundInList = true;
						urlsCount[j]++;
						break;
					}
				}
				if(!foundInList) {
					// Not in list, add the current url:
					urls[j] = url;
					urlsCount[j] = 1;
				}
			}
			
			// Add the labeltext plus shortcut, if not empty:
			var labelText = this.getStringBundle().getString('tooltipLogin');
			var formattedShortcut = this.getFormattedShortcut();
			if(formattedShortcut)
				labelText += ' (' + this.getFormattedShortcut() + ')';

			label.setAttribute('class',	'bold');
			label.setAttribute(
				'value',
				labelText
			);
			tooltip.appendChild(label);
			
			// Add a description of the URL elements and count:
			hbox = document.createElement('hbox');		
			descr = document.createElement('description');
			descr.setAttribute('class', 'italic');
			descr.setAttribute(
				'value',
				this.getStringBundle().getString('tooltipLoginUrl')
			);
			hbox.appendChild(descr);
			spacer = document.createElement('spacer');
			spacer.setAttribute('flex','1');
			hbox.appendChild(spacer);
			descr = document.createElement('description');
			descr.setAttribute('class', 'italic');
			descr.setAttribute(
				'value',
				this.getStringBundle().getString('tooltipLoginUrlCount')
			);
			hbox.appendChild(descr);
			tooltip.appendChild(hbox)
			
			// Add the url list:
			for(var i = 0; i < urls.length; i++) {
				descr = document.createElement('description');
				descr.setAttribute(
					'value',
					urls[i] + '  (' + urlsCount[i] + ')'
				);
				tooltip.appendChild(descr);
			}
		} else {
			label.setAttribute(
				'value',
				this.getStringBundle().getString('tooltipNoLogin')
			);
			tooltip.appendChild(label);
		}
	},

	contextMenu: function(event) {
		this.menuPreparation('secureLoginBookmarkContextItem', 'secureLoginContextAutofillFormsMenu');
	},

	toolsMenu: function(event) {
		this.menuPreparation('secureLoginBookmarkToolsMenuItem', 'secureLoginToolsMenuAutofillFormsMenu');
	},
	
	buttonMenu: function(event) {
		this.menuPreparation('secureLoginBookmarkButtonMenuItem', 'secureLoginButtonMenuAutofillFormsMenu');
	},	

	menuPreparation: function(bookmarkItemID, autofillFormsMenuID) {
		var bookmarkItem = document.getElementById(bookmarkItemID);
		if(bookmarkItem) {
			if(this.getSecureLoginPrefs().getBoolPref('secureLoginBookmarks') &&
				this.getDoc().location && (this.getDoc().location.protocol == 'http:' || this.getDoc().location.protocol == 'https:')) {
				bookmarkItem.setAttribute('disabled', 'false');
			} else {
				bookmarkItem.setAttribute('disabled', 'true');
			}
		}
		var autofillFormsPopupMenu = document.getElementById('autofillFormsPopupMenu');
		var autofillFormsMenu = document.getElementById(autofillFormsMenuID);
		var autofillFormsMenuSeparator = document.getElementById(autofillFormsMenuID+'Separator');
		if(this.getSecureLoginPrefs().getBoolPref('autofillFormsOnLogin') && autofillFormsPopupMenu) {
			if(autofillFormsMenu && !autofillFormsMenu.hasChildNodes()) {
				autofillFormsPopupMenu = autofillFormsPopupMenu.cloneNode(true);
				autofillFormsPopupMenu.removeAttribute('position');
				autofillFormsMenu.appendChild(autofillFormsPopupMenu);
			}
			if(autofillFormsMenu) {
				autofillFormsMenu.removeAttribute('hidden');
			}
			if(autofillFormsMenuSeparator) {
				autofillFormsMenuSeparator.removeAttribute('hidden');	
			}
		} else {
			if(autofillFormsMenu) {
				autofillFormsMenu.setAttribute('hidden', 'true');
			}
			if(autofillFormsMenuSeparator) {
				autofillFormsMenuSeparator.setAttribute('hidden', 'true');
			}
		}
	},
	
	clickHandler: function(event) {
		switch(event.button) {
			case 0:
				if(event.target.id == 'secureLoginPanelIcon') {
					// The left mouse button already performs the login command for the secureLoginButton,
					// but not for the status bar icon:
					this.userSelectionLogin(event);
				}
				break;
			case 1:
				this.masterSecurityDeviceLogout(event);
				break;
		}
	},

	getMasterSecurityDevice: function() {
		return Components.classes['@mozilla.org/security/pk11tokendb;1']
				.getService(Components.interfaces.nsIPK11TokenDB);
	},
	
	masterSecurityDeviceLogout: function(event) {	
		if(this.getMasterSecurityDevice().getInternalKeyToken().isLoggedIn()) {
			this.getMasterSecurityDevice().findTokenByName('').logoutAndDropAuthenticatedResources();
		}
		this.showAndRemoveNotification(this.getStringBundle().getString('masterSecurityDeviceLogout'));
	},

	showAndRemoveNotification: function(label, timeout, id, image, priority, buttons) {
		timeout = timeout ? timeout : this.getSecureLoginPrefs().getIntPref('defaultNotificationTimeout');
		id = id ? id : 'secureLoginNotification';
		image = image ? image : 'chrome://secureLogin/skin/login-small.png';
		priority = priority ? priority : 'PRIORITY_INFO_HIGH';
		buttons = buttons ? buttons : null;
		this.showNotification(label, id, image, priority, buttons);
		// Automatically remove the notification after the timeout:
		window.setTimeout(
			function() {
				secureLogin.removeNotification()
			},
			timeout
		);
	},
	
	showNotification: function(label, id, image, priority, buttons) {
		id = id ? id : 'secureLoginNotification';
		image = image ? image : 'chrome://secureLogin/skin/login-small.png';
		priority = priority ? priority : 'PRIORITY_INFO_HIGH';
		buttons = buttons ? buttons : null;
		// First remove notifications with the same id:
		this.removeNotification(id);
		var notificationBox = this.getBrowser().getNotificationBox();
		if(notificationBox) {
			notificationBox.appendNotification(
				label,
				id,
				image,
				priority,
				buttons
			);			
		}
	},
	
	removeNotification: function(id) {
		id = id ? id : 'secureLoginNotification';
		var notificationBox = this.getBrowser().getNotificationBox();
		if(notificationBox) {
			var notification = notificationBox.getNotificationWithValue(id);
			if(notification) {
				notificationBox.removeNotification(notification);				
			}
		}
	},

	contextMenuSelectionLogin: function(popup) {
		try {
			if(!this.secureLogins[0]) {
				// On Firefox 3 we still have to get the valid login objects:
				this.secureLogins = this.getRealLoginObjects();
				
				// Return if the list of login objects is empty (should not happen):
				if(!this.secureLogins || this.secureLogins.length == 0) {
					return false;
				}
			}
			this.prepareUserSelectionPopup(popup);
		} catch (e) {
   			// Decrypting failed
     		return false;
  		}
	},
	
	prepareUserSelectionPopup: function(popup) {
		// Remove the old child nodes (should be already removed by the popuphiding event):
		while(popup.hasChildNodes()) {
			popup.removeChild(popup.firstChild);
		}
		
		var menuitem = document.createElement('menuitem');
		menuitem.setAttribute('class','menuitem-iconic');
		menuitem.setAttribute('image','chrome://secureLogin/skin/user-small.png');
		// Add a menuitem for each available user login:
		for(var i = 0; i < this.secureLogins.length; i++) {
			var username = this.getUsernameFromLoginObject(this.secureLogins[i]);
			// Show form index?
			if(this.showFormIndex) {
				username += '  (' + this.secureLoginsFormIndex[i] + ')';
			}
			menuitem = menuitem.cloneNode(false);
			menuitem.setAttribute('label',username);
			menuitem.setAttribute('oncommand','secureLogin.login(null, '+i+', true);');
			popup.appendChild(menuitem);
		}
	},
	
	userSelectionLogin: function(event) {		
		// Search for valid logins and outline login fields if not done automatically:
		if(!this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload')) {
			this.searchLoginsInitialize();
		}
		
		// Check for valid logins:
		if(this.secureLogins && this.secureLogins.length > 0) {
			if(this.secureLogins.length > 1) {				
				// Determine if no master password is set or the user has already been authenticated:
				var masterPasswordRequired = true;
				if(!this.getMasterSecurityDevice().getInternalKeyToken().needsLogin()
					|| this.getMasterSecurityDevice().getInternalKeyToken().isLoggedIn()) {
					masterPasswordRequired = false;
				}
				var popup = document.getElementById('secureLoginUserSelectionPopup');			
				if(popup && typeof popup.openPopup == 'function' && !masterPasswordRequired) {
					try {
						if(!this.secureLogins[0]) {
							// On Firefox 3 we still have to get the valid login objects:
							this.secureLogins = this.getRealLoginObjects();
							
							// Return if the list of login objects is empty (should not happen):
							if(!this.secureLogins || this.secureLogins.length == 0) {
								return;
							}
						}
						this.prepareUserSelectionPopup(popup);					
						// Show the popup menu (only available for Firefox >= 3):
						popup.openPopup(event.target, null, 0, 0, false, true);
					} catch (e) {
			   			// Decrypting failed
			     		return;
			  		}
				} else {
					// Show a selection box instead of the popup menu:
					this.login(null, null, true);
				}
			} else {
				// Just login with the single available username:
				this.login(null, 0, true);
			}
		} else {	
			// Autofill Forms integration (requires extension autofillForms@blueimp.net):
			if(this.getSecureLoginPrefs().getBoolPref('autofillFormsOnLogin')) {
				try {
					autofillForms.fillForms();
				} catch(e) {
					alert(e);
				}
			}
		}
	},
	
	login: function(win, loginIndex, skipLoginSearch) {
		if(!win || !win.document) {
			win = this.getWin();
		}
		
		// Autofill Forms integration (requires extension autofillForms@blueimp.net):
		if(this.getSecureLoginPrefs().getBoolPref('autofillFormsOnLogin')) {
			try {
				autofillForms.fillForms(win);
			} catch(e) {
				//alert(e);
			}
		}
		
		// Search for valid logins and outline login fields if not done automatically:
		if(!this.getSecureLoginPrefs().getBoolPref('searchLoginsOnload') && !skipLoginSearch) {
			this.searchLoginsInitialize(win);
		}
	   	   
		// Check for valid logins:
		if(this.secureLogins && this.secureLogins.length > 0) {
			try {
				if(!this.secureLogins[0]) {
					// On Firefox 3 we still have to get the valid login objects:
					this.secureLogins = this.getRealLoginObjects();
					
					// Return if the list of login objects is empty (user canceled master password entry):
					if(!this.secureLogins || this.secureLogins.length == 0) {
						return;
					}
				}
				
				// The list index of the login:
				var selectedIndex = 0;
				
				// Prompt for a selection, if list contains more than one login:
				if(this.secureLogins.length > 1) {			
					// Check if the loginIndex contains an index to select:
					if(typeof loginIndex != 'undefined' && !isNaN(parseInt(loginIndex)) && loginIndex < this.secureLogins.length) {
						selectedIndex = loginIndex;
					} else {
						var list = new Array(this.secureLogins.length);
						for(var i = 0; i < this.secureLogins.length; i++) {
							list[i] = this.getUsernameFromLoginObject(this.secureLogins[i]);
							// Show form index?
							if(this.showFormIndex) {
								list[i] += '  (' + this.secureLoginsFormIndex[i] + ')';
							}
						}
						var selected = {};
						
						var selectionPrompt = this.getStringBundle().getString('loginSelectionPrompt');
						if(this.showFormIndex)
							selectionPrompt += '  (' + this.getStringBundle().getString('formIndex') + ')';
						
						var ok = this.getPrompts().select(
							window,
							this.getStringBundle().getString('loginSelectionWindowTitle'),
							selectionPrompt + ':',
							list.length,
							list,
							selected
						);
						
						if(ok == false)
							return;
							
						// Set the list index to the selected one:
						selectedIndex = selected.value
					}
				}
				
				// Set the win object to the window (frame) containing the login form:
				win = this.secureLoginsWindow[selectedIndex];
				
				// The index for the form containing the login fields:
				var formIndex = this.secureLoginsFormIndex[selectedIndex];
				
				// The form elements list:
				var elements = this.getDoc(win).forms[formIndex].elements;
				
				// User + Pass fields:
				var usernameField = this.secureLoginsUserField[selectedIndex];
				var passwordField = this.secureLoginsPassField[selectedIndex];	
				
				// The charset of the given document:
				var charset = this.getDoc(win).characterSet;
				
				// Get the target url from the form action value or if empty from the current document:			
			    var url = this.getDoc(win).forms[formIndex].action ? this.getDoc(win).forms[formIndex].action : this.getDoc(win).baseURI;
			    
				// Ask for confirmation if we had a failed bookmark-login:	    
			    if(this.failedBookmarkLogin) {
					var continueLogin = this.getPrompts().confirm(
						null,
						this.getStringBundle().getString('domainChangeConfirmTitle'),
						this.getStringBundle().getString('domainChangeConfirmDesc') + ' ' + url
					);
					if(!continueLogin)
						return;
				} else if(this.getSecureLoginPrefs().getBoolPref('domainChangeConfirm')) {
					// Ask for confirmation, if (second level) domain in target url is different from current document:
			    	var domainChange = true;
				    try{
				    	// Get source and target host:
				    	var targetHost = this.makeURI(url, charset).host;
				    	var sourceHost = this.getDoc(win).location.hostname;
						
				    	// Check if host is not changing at all (same IP or domain):
				    	if(sourceHost == targetHost)
				    		domainChange = false;
				    		
				    	// Check if second level domain is changing:
				    	else {							
							// Extract the second level domain (e.g. example.org from secure.example.org) from source host:
			    			var sld = sourceHost.match(/[\w-]+\.\w+$/)[0];
			    			
			    			// Search the targetHost for the extracted second level domain:
			    			var sldIndex = targetHost.lastIndexOf(sld);

			    			// Check if the sourceHost sld is the same as the target host sld:
			    			if(sld && (sldIndex != -1) && (targetHost.length == (sldIndex + sld.length)))
				    			domainChange = false;
				    	}
				    } catch (e) {
				    	// Target url is invalid
				    }
				    if(domainChange) {
						var checkConfirm = {value: true};
						var continueLogin = this.getPrompts().confirmCheck(
							null,
							this.getStringBundle().getString('domainChangeConfirmTitle'),
							this.getStringBundle().getString('domainChangeConfirmDesc') + ' ' + url,
							this.getStringBundle().getString('domainChangeConfirmPrefLabel'),
							checkConfirm
						);
						if(!continueLogin)
							return;
						else if(!checkConfirm.value)
							this.getSecureLoginPrefs().setBoolPref('domainChangeConfirm', false);
				    }
			    }
				
				// Reset failed bookmark-login:
				this.failedBookmarkLogin = null;
				
				// If JavaScript protection is to be used, check the exception list:
				var useJavaScriptProtection = this.getSecureLoginPrefs().getBoolPref('javascriptProtection');
				if(useJavaScriptProtection && this.inArray(this.getExceptions(), this.getDoc(win).location.protocol + '//' + this.getDoc(win).location.host))
					useJavaScriptProtection = false;
				
			    // Send login data without using the form:
			    if(useJavaScriptProtection) {
					    	
			    	// String to save the form data:
			    	var dataString = '';
			    	
			    	// Local helper function to add name and value pairs urlEncoded to the dataString:
			    	function addToDataString(parentObject, name, value) {
			    		if(dataString)
			    			dataString += '&';
			    		dataString += parentObject.urlEncode(name, charset)
			    					+ '='
			    					+ parentObject.urlEncode(value, charset);
			    	}

			    	var submitButtonFound = false;
			    	
			    	// Search for form elements other than user+pass fields and add them to the dataString:
			    	for(var i = 0; i < elements.length; i++) {
			    		
			    		// Don't add disabled elements or elements without a "name":
			    		if(!elements[i].name || elements[i].disabled)
			    			continue;
						
						switch(elements[i].type) {
							case 'hidden':
								addToDataString(this, elements[i].name, elements[i].value);
								break;
							case 'text':
								if(!usernameField || elements[i].name != usernameField.name)
		    						addToDataString(this, elements[i].name, elements[i].value);
								break;
							case 'select-one':
								addToDataString(this, elements[i].name, elements[i].value);
								break;
							case 'select-multiple':
								for(var j = 0; j < elements[i].options.length; j++) {
									if(elements[i].options[j].selected)
										addToDataString(this, elements[i].name, elements[i].options[j].value);
								}
								break;
							case 'checkbox':
								if(elements[i].checked)
		    						addToDataString(this, elements[i].name, elements[i].value);
								break;
							case 'radio':
								if(elements[i].checked)
		    						addToDataString(this, elements[i].name, elements[i].value);
								break;
							case 'submit':
								// Only add first submit button:
								if(!submitButtonFound) {
									addToDataString(this, elements[i].name, elements[i].value);
									submitButtonFound = true;
								}
								break;
						}

			    	}
			    	
			    	// If no submit button found, search for an input of type="image" which ist not in the elements list:
			    	if(!submitButtonFound) {
				     	var inputElements = this.getDoc(win).forms[formIndex].getElementsByTagName('input');
				     	for(var i = 0; i < inputElements.length; i++) {
							if(inputElements[i].type == 'image') {
								// Image submit buttons add the "click-coordinates" name.x and name.y to the request data:
								addToDataString(this, inputElements[i].name + '.x', 1);
								addToDataString(this, inputElements[i].name + '.y', 1);
								addToDataString(this, inputElements[i].name, inputElements[i].value);
							}
				     	}
			     	}
			     	
			     	// Add user and pass to the dataString using the saved login data:
			     	if(usernameField)
			     		addToDataString(this, usernameField.name, this.getUsernameFromLoginObject(this.secureLogins[selectedIndex]));
			     	addToDataString(this, passwordField.name, this.getPasswordFromLoginObject(this.secureLogins[selectedIndex]));
			     	
			     	// Check if the url is a valid one (HTTP or HTTPS):
			     	if(url.indexOf('http://') != 0 && url.indexOf('https://') != 0)
			     		return;
			     	
			     	// Check if the url is an allowed one (throws an exception if not):
					this.urlSecurityCheck(url, this.getDoc(win).location.href);
			     	
			     	// Send the data by GET or POST:
			     	if(this.getDoc(win).forms[formIndex].method && this.getDoc(win).forms[formIndex].method.toLowerCase() == 'get') {
			     		// Add the parameter list to the url, remove existing parameters:
			     		var paramIndex = url.indexOf('?');
			     		if(paramIndex == -1)
			     			url += '?' + dataString;
			     		else
			     			url = url.substring(0, paramIndex+1) + dataString;			     		
			     		// Load the url in the current window (params are url, referrer and post data):
						loadURI(url, this.makeURI(this.getDoc(win).location.href, charset), null);
			     	} else {
			     		// Create post data mime stream (params are aStringData, aKeyword, aEncKeyword, aType):
						var postData = getPostDataStream(dataString, '', '', 'application/x-www-form-urlencoded');
						// Load the url in the current window (params are url, referrer and post data):
						loadURI(url, this.makeURI(this.getDoc(win).location.href, charset), postData);
			     	}
			    	
				} else {
				    // Fill the login fields:
				    if(usernameField)
						usernameField.value = this.getUsernameFromLoginObject(this.secureLogins[selectedIndex]);
					passwordField.value = this.getPasswordFromLoginObject(this.secureLogins[selectedIndex]);
			     	
			     	if(this.getSecureLoginPrefs().getBoolPref('autoSubmitForm')) {
				     	// Prevent multiple submits (e.g. if submit is delayed) by setting a variable (after click on a submit button):
				     	var submitted = false;
			     		// Search for the submit button:
				     	for(var i = 0; i < elements.length; i++) {
				     		// auto-login by clicking on the submit button:
							if(elements[i].type && elements[i].type == 'submit') {
								elements[i].click();
								submitted = true;
								break;
							}
				     	}

				     	if(!submitted) {
						    // Search for a submit button of type="image" which ist not in the elements list:
					     	var inputElements = this.getDoc(win).getElementsByTagName('input');
					     	for(var i = 0; i < inputElements.length; i++) {
					     		// auto-login by clicking on the image submit button if it belongs to the current form:
								if(inputElements[i].type == 'image' && inputElements[i].form && inputElements[i].form == this.getDoc(win).forms[formIndex]) {
									inputElements[i].click();
									submitted = true;
									break;
								}
					     	}

					     	if(!submitted) {
						     	// No submit button found, try to submit anyway:
						     	this.getDoc(win).forms[formIndex].submit();
					     	}
				     	}
			     	} else {
			     		// Don't submit automatically but set the focus on the password field,
			     		// this way submitting can be done by hitting return on the keyboard:
			     		passwordField.focus();
			     		return;
			     	}
			    }
			    
			    // Play sound notification:
		     	if(this.getSecureLoginPrefs().getBoolPref('playLoginSound'))
		     		this.playSound('loginSoundFileName');
									
			} catch (e) {
	   			// Decrypting failed or url is not allowed
	   			// alert(e);
	     		return;
	  		}
		}
		
		// Reset secure login objects to release memory:
		this.secureLogins = null;
		this.secureLoginsFormIndex = null;
		this.secureLoginsPassField = null;
		this.secureLoginsUserField = null;
		this.secureLoginsWindow = null;
	},
	
	getUsernameFromLoginObject: function(loginObject) {
		if(this.getLoginManager()) {
			// Firefox 3:
			return loginObject.username;
		} else {
			// Versions before Firefox 3:
			return loginObject.user;
		}
	},
	
	getPasswordFromLoginObject: function(loginObject) {
		// Both login objects (Firefox 3 and before) contain a "password" attribute:
		return loginObject.password;
	},
	
	getExceptions: function() {
		// Get the exception list from the preferences:
		var exceptions = this.getSecureLoginPrefs()
							.getComplexValue('exceptionList',Components.interfaces.nsISupportsString)
							.data.split(' ');
		return exceptions && exceptions[0] ? exceptions : new Array();
	},

	setExceptions: function(exceptions) {
		// Store the exceptions separated by spaces as unicode string in the preferences:
		this.getSecureLoginPrefs().setComplexValue(
			'exceptionList',
			Components.interfaces.nsISupportsString,
			this.getUnicodeString(exceptions.join(' '))
		);
	},

	recognizeKeys: function(doc, event) {
		var modifiers = new Array();
		var key = '';
		var keycode = '';
		
		// Get the modifiers:
		if(event.altKey) modifiers.push('alt');
		if(event.ctrlKey) modifiers.push('control');
		if(event.metaKey) modifiers.push('meta');
		if(event.shiftKey) modifiers.push('shift');
		
		// Get the key or keycode:
		if(event.charCode) {
			key = String.fromCharCode(event.charCode).toUpperCase();
		} else {
			// Get the keycode from the keycodes list:
			keycode = this.getKeyCodes()[event.keyCode];
			if(!keycode)
				return;
		}
		
		// Shortcut may be anything, but not 'VK_TAB' alone (without modifiers),
		// as this button is used to change focus to the 'Apply' button: 
		if(modifiers.length > 0 || keycode != 'VK_TAB') {
			this.shortcutNew = new Object();
			this.shortcutNew['modifiers'] 	= modifiers;
			this.shortcutNew['key'] 		= key;
			this.shortcutNew['keycode'] 	= keycode;
			doc.getElementById(event.target.id).value = this.getFormattedShortcut(this.shortcutNew);
		}
	},
		
	getKeyCodes: function() {
		var keycodes = new Array();
		// Get the list of keycodes from the KeyEvent object:
		for(var property in KeyEvent) {
			keycodes[KeyEvent[property]] = property.replace('DOM_','');
		}
		// VK_BACK_SPACE (index 8) must be VK_BACK:
		keycodes[8] = 'VK_BACK';
		return keycodes;
	},

	applyShortcut: function(doc, event) {
		// Apply only if a new shortcut is available:
		if(!this.shortcutNew)
			return;
		this.setShortcut(this.shortcutNew);
	},
	
	disableShortcut: function(doc, event) {
		// Disables the shortcut:
		this.setShortcut(null);
		doc.getElementById('keyboardShortcut').value = '';
		this.shortcutNew = null;
	},
	
	discardShortcut: function() {
		// Discard the shortcut changes:
		this.shortcutNew = null;
	},
	
	getShortcut: function() {
		if(this.shortcut == null) {
			// Create a new shortcut object:
			this.shortcut = new Object();
			var shortcutItems = this.getSecureLoginPrefs()
								.getComplexValue('shortcut',Components.interfaces.nsIPrefLocalizedString)
								.data.split('+');
			if(shortcutItems.length == 0) {
				this.shortcut['key'] 		= '';
				this.shortcut['keycode'] 	= '';
			} else {
				// Remove the last element and save it as key:
				var key = shortcutItems.pop();
				// Check if the key is a keycode:
				if(key.indexOf('VK') == 0) {
					this.shortcut['key'] 		= '';
					this.shortcut['keycode'] 	= key;
				} else {
					this.shortcut['key'] 		= key;
					this.shortcut['keycode'] 	= '';
				}
			}
			// Save the remaining items as modifiers:
			this.shortcut['modifiers'] 	= shortcutItems;
		}
		return this.shortcut;
	},
	
	setShortcut: function(shortcut) {
		var stringData = '';
		if(shortcut) {
			if(shortcut['modifiers'] && shortcut['modifiers'].length > 0)
				// Add the modifiers, separated by '+' from each other and from the key:
				stringData += shortcut['modifiers'].join('+') + '+';
			if(shortcut['key'])
				// Add the key:
				stringData += shortcut['key'];
			else if(shortcut['keycode'])
				// Add the keycode (instead of a key):
				stringData += shortcut['keycode'];
		}
		// Save the shortcut as Unicode String in the preferences:
		this.getSecureLoginPrefs().setComplexValue(
			'shortcut',
			Components.interfaces.nsISupportsString,
			this.getUnicodeString(stringData)
		);
	},
	
	updateShortcut: function() {
		// Setting the shortcut object to "null" will update it on the next getShortcut() call:
		this.shortcut = null;
		// Get the keyboard shortcut elements:
		var modifiers = this.getShortcut()['modifiers'].join(' ');
		var key = this.getShortcut()['key'];
		var keycode = this.getShortcut()['keycode'];

		// Remove current key if existing:
		if(document.getElementById('secureLoginShortCut')) {
			document.getElementById('mainKeyset').removeChild(
				document.getElementById('secureLoginShortCut')
			);
		}
		
		// Check if keyboard shortcut is enabled (either key or keycode set):
		if(key || keycode) {
			// Create a key element:
			var keyNode = document.createElement('key');
			
			keyNode.setAttribute('id', 'secureLoginShortCut');
			keyNode.setAttribute('command', 'secureLogin');
			
			// Set the key attributes from saved shortcut:
			keyNode.setAttribute('modifiers', modifiers);
			if(key) {
				keyNode.setAttribute('key', key);
			} else {
				keyNode.setAttribute('keycode', keycode);
			}
			
			// Add the key to the mainKeyset:
			document.getElementById('mainKeyset').appendChild(keyNode);
		}
	},
	
	getFormattedShortcut: function(shortcutParam) {
		// Get shortcut from param or take the object attribute:
		var shortcut = shortcutParam ? shortcutParam : this.getShortcut();
		var formattedShortcut = '';
		// Add the modifiers:
		for(var i=0; i < shortcut['modifiers'].length; i++)
			try {
				formattedShortcut += this.getStringBundle().getString(shortcut['modifiers'][i]) + '+';
			} catch(e) {
				// Error in shortcut string, return empty String;
				return '';
			}
		if(shortcut['key'])
			// Add the key:
			if(shortcut['key'] == ' ')
				formattedShortcut += this.getStringBundle().getString('VK_SPACE');
			else
				formattedShortcut += shortcut['key'];
		else if(shortcut['keycode'])
			// Add the keycode (instead of the key):
			try {
				formattedShortcut += this.getStringBundle().getString(shortcut['keycode']);
			} catch(e) {
				formattedShortcut += shortcut['keycode'].replace('VK_', '');
			}
		return formattedShortcut;
	},
	
	selectAudioFile: function(doc, prefName) {
		// doc is the current document from which the method has been called
		// prefName is the preference name as well as the textbox id
		
		try {        
			// Create a file picker instance:
			var fp = Components.classes['@mozilla.org/filepicker;1']
						.createInstance(Components.interfaces.nsIFilePicker);
			
			// Initialize the file picker window:
			fp.init(
				window,
				this.getStringBundle().getString('selectAudioFile'),
				Components.interfaces.nsIFilePicker.modeOpen
			);
			
			// Apply a file filter for wave files:
			fp.appendFilter('*.wav','*.wav;*.WAV');
			fp.filterIndex=0;
			
			// Show the file picker window:
			var rv = fp.show();
			
			if (rv == Components.interfaces.nsIFilePicker.returnOK) {
				var file = fp.file;
				// Save the selected file in the preferences:
				this.getSecureLoginPrefs().setComplexValue(prefName, Components.interfaces.nsILocalFile, file);
				// Save the selected file in the associated textbox:
				doc.getElementById(prefName).value = file.path;
			}
		} catch(e) {
			// alert(e);
		}
	},

	playSound: function(prefName) {
        try {
        	// Get the filename stored in the preferences:
	        var file = this.getSecureLoginPrefs().getComplexValue(prefName, Components.interfaces.nsILocalFile);
	        
	        // Get an url for the file:
	        var url = this.getIOS().newFileURI(file, null, null);
	        
	        // Play the sound:
	        this.getSound().play(url);
		} catch(ex) {
			// No file found
		}
	},

	showDialog: function(url) {
		window.openDialog(
			url,
			'',
			'chrome=yes,resizable=yes,toolbar=yes,centerscreen=yes,modal=no,dependent=no,dialog=no',
			this
		);
	},

	showBookmarkDialog: function() {
		if(this.getDoc().location && (this.getDoc().location.protocol == 'http:' || this.getDoc().location.protocol == 'https:')) {
			var url;
			// Create a Secure Login Bookmark out of the current URL:
			if(this.getDoc().location.hash) {
				var regExp = new RegExp(this.getDoc().location.hash + '$');
				url = this.getDoc().location.href.replace(regExp, this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash'));
			} else {
				url = this.getDoc().location.href + this.getSecureLoginPrefs().getCharPref('secureLoginBookmarkHash');
			}

			if(this.getVersionComparator().compare(this.getAppInfo().version, '2.*') < 0) {
				// Firefox version 1.5 - 2.0.0.*:
				var bookmarkArguments = {
					name: this.getDoc().title,
					url: url,
					charset: this.getDoc().characterSet
				}
				window.openDialog(
					'chrome://browser/content/bookmarks/addBookmark2.xul',
					'', 
					'centerscreen=yes,chrome=yes,dialog=yes,resizable=yes,dependent=yes',
					bookmarkArguments
				);
			} else {
				// Firefox version 3 and prereleases:
				var bookmarkArguments = {
					action: 'add',
					type: 'bookmark',
					hiddenRows: ['location', 'description', 'load in sidebar'],
					uri: this.makeURI(url, this.getDoc().characterSet),
					title: this.getDoc().title
				};
				window.openDialog(
					'chrome://browser/content/places/bookmarkProperties2.xul',
					'', 
					'centerscreen=yes,chrome=yes,dialog=yes,resizable=yes,dependent=yes',
					bookmarkArguments
				);
			}
		}
	},

	urlSecurityCheck: function(url, sourceURL) {
		try {
			this.getSecManager().checkLoadURIStr(sourceURL, url, Components.interfaces.nsIScriptSecurityManager.STANDARD);
		} catch (e) {
			throw 'Loading of ' + url + ' denied.';
		}
	},

	makeURI: function(aURL, aOriginCharset) {
		return this.getIOS().newURI(aURL, aOriginCharset, null);
	},
	
	urlEncode: function(string, charset) {
		if(charset == 'UTF-8') {
			// encodeURIComponent encodes the strings by using escape sequences
			// representing the UTF-8 encoding of the character:
			return encodeURIComponent(string);
		} else {
			// This escapes characters representing the given charset,
			// it won't work if the given string is not part of the charset
			return this.getTextToSubURI().ConvertAndEscape(charset, string);
		}
	},

	getTextToSubURI: function() {
		return Components.classes['@mozilla.org/intl/texttosuburi;1']
				.getService(Components.interfaces.nsITextToSubURI);
	},
	
	getUnicodeString: function(stringData) {
		// Create an Unicode String:
		var str = Components.classes['@mozilla.org/supports-string;1']
					.createInstance(Components.interfaces.nsISupportsString);
		// Set the String value:
		str.data = stringData;
		// Return the Unicode String:
		return str;
	},

	getStringBundle: function() {
		return document.getElementById('secureLoginStringBundle');
	},
	
	getDoc: function(win) {
		if(win)
			return win.document;
		else if(content)
			return content.document;
		else
			return this.getBrowser().contentDocument;
	},
	
	getWin: function() {
		if(content)
			return content;
		else
			return this.getBrowser().contentWindow;
	},

	getBrowser: function() {
		return this.getWindowMediator().getMostRecentWindow('navigator:browser').getBrowser();
	},

	getWindowMediator: function() {
		return Components.classes['@mozilla.org/appshell/window-mediator;1']
				.getService(Components.interfaces.nsIWindowMediator);
	},
	
	getPasswordManager: function() {
		// PasswordManager doesn't exist in Firefox 3:
		if(!Components.classes['@mozilla.org/passwordmanager;1'])
			return null;
		return Components.classes['@mozilla.org/passwordmanager;1']
				.getService(Components.interfaces.nsIPasswordManager);
	},

	getLoginManager: function() {
		// LoginManager only exists in Firefox 3:
		if(!Components.classes['@mozilla.org/login-manager;1'])
			return null;
		return Components.classes['@mozilla.org/login-manager;1']
				.getService(Components.interfaces.nsILoginManager);
	},
	
	getPrefManager: function() {
		return Components.classes['@mozilla.org/preferences-service;1']
				.getService(Components.interfaces.nsIPrefService);
	},
	
	getSecManager: function() {
		return Components.classes['@mozilla.org/scriptsecuritymanager;1']
				.getService(Components.interfaces.nsIScriptSecurityManager);
	},
	
	getIOS: function() {
		return Components.classes['@mozilla.org/network/io-service;1']
				.getService(Components.interfaces.nsIIOService);
	},
	
	getFileHandler: function() {
		return this.getIOS().getProtocolHandler('file')
				.QueryInterface(Components.interfaces.nsIFileProtocolHandler);
	},
	
	getSecureLoginPrefs: function() {
		if(this.secureLoginPrefs == null)
			this.secureLoginPrefs = this.getPrefManager().getBranch('extensions.secureLogin@blueimp.net.');
		return this.secureLoginPrefs;
	},
	
	getSound: function() {
		return Components.classes['@mozilla.org/sound;1']
				.createInstance(Components.interfaces.nsISound);
	},
	
	getPrompts: function() {
		return Components.classes['@mozilla.org/embedcomp/prompt-service;1']
				.getService(Components.interfaces.nsIPromptService);
	},
	
	getAppInfo: function() {
		return Components.classes['@mozilla.org/xre/app-info;1']
				.getService(Components.interfaces.nsIXULAppInfo);
	},
	
	getVersionComparator: function() {
		return Components.classes['@mozilla.org/xpcom/version-comparator;1']
				.getService(Components.interfaces.nsIVersionComparator);
	},
		
	inArray: function(array, item) {
		var i = array.length;
		while(i--)
			if(array[i] === item)
				return true;
		return false;
	},

	openHelp: function(topic) {
		if(!topic) {
			topic = '';
		}
		var url = this.getSecureLoginPrefs().getCharPref('helpURL').replace(/\[TOPIC\]$/, topic);
		this.openNewTab(url, true);
	},
	
	openNewTab: function(url, focus) {
		var helpTab = this.getBrowser().addTab(url);
		if(focus) {
			this.getBrowser().selectedTab = helpTab;
			this.getWindowMediator().getMostRecentWindow('navigator:browser').focus();
		}
	},

	log: function(aMessage, aSourceName, aSourceLine, aLineNumber, aColumnNumber, aFlags, aCategory) {
		var consoleService = Components.classes['@mozilla.org/consoleservice;1']
			.getService(Components.interfaces.nsIConsoleService);
		if(aSourceName != 'undefined') {
			var scriptError = Components.classes["@mozilla.org/scripterror;1"]
				.createInstance(Components.interfaces.nsIScriptError);
			scriptError.init(
				aMessage,
				aSourceName,
				aSourceLine,
				aLineNumber,
				aColumnNumber,
				aFlags,
				aCategory
			);
			consoleService.logMessage(scriptError);
		} else {
			consoleService.logStringMessage(aMessage);	
		}
	},

	optionsInit: function() {
		// Display the shortcut combination:
		document.getElementById('keyboardShortcut').value = this.getFormattedShortcut();
		
		// Display the filenames stored in the preferences:
		var file;
		try {
			file = this.getSecureLoginPrefs().getComplexValue('loginFoundSoundFileName', Components.interfaces.nsILocalFile);
			document.getElementById('loginFoundSoundFileName').value = file.path;
		} catch(e) {
			// alert(e);
			// No file found
		}
		try {
			file = this.getSecureLoginPrefs().getComplexValue('loginSoundFileName', Components.interfaces.nsILocalFile);
			document.getElementById('loginSoundFileName').value = file.path;
		} catch(e) {
			// alert(e);
			// No file found
		}
		
		// Adjust the CSS definitions for Firefox < 2.0 so the options icons are displayed properly:
		if(this.getVersionComparator().compare(this.getAppInfo().version, '2.0') < 0) {
			for(var i=0; i<document.styleSheets[1].cssRules.length; i++) {
					if(document.styleSheets[1].cssRules[i].selectorText == '*|#secureLoginPrefs *|radio[pane="secureLoginPrefPaneAdvanced"]')
						document.styleSheets[1].cssRules[i].style.cssText = '-moz-image-region: rect(0px, 160px, 32px, 128px);';
					else if(document.styleSheets[1].cssRules[i].selectorText == '*|#secureLoginPrefs *|radio[pane="secureLoginPrefPaneGUI"]')
						document.styleSheets[1].cssRules[i].style.cssText = '-moz-image-region: rect(0px, 192px, 32px, 160px);';
			}
		}
	},
	
	optionsDestroy: function() {
		this.discardShortcut();
	},

	exceptionsInit: function() {
		// Copy the secureLogin exception array into the local list:
		this.exceptions = this.getExceptions().slice();
				
		// Get the tree:
		this.exceptionsTree = document.getElementById('exceptionsTree');
		
		// Implement the TreeView interface:
		this.exceptionsTreeView = {
			rowCount: 0,
			setTree: function(tree){},
			getImageSrc: function(row,column) {},
			getProgressMode: function(row,column) {},
			getCellValue: function(row,column) {},
			getCellText: function(row,column){
				if(column.id=='exceptionsCol')
					return this.parent.exceptions[row];
				else
					return '';
			},
			isSeparator: function(index) {return false;},
			isSorted: function() { return false; },
			isContainer: function(index) {return false;},
			cycleHeader: function(column) {},
			getRowProperties: function(row,prop){},
			getColumnProperties: function(column,prop){},
			getCellProperties: function(row,column,prop){},
			getParentIndex: function(index) {return -1}
		};
		// Set the secureLogin object as parent:
		this.exceptionsTreeView.parent = this;
		
		// Set the tree length using the exception list length:
		this.exceptionsTreeView.rowCount = this.exceptions.length;
		
		// Enable the "removeAllButton" if exceptions are stored:
		if(this.exceptionsTreeView.rowCount > 0)
			document.getElementById('removeAllButton').setAttribute('disabled', 'false');
		
		// If the current url is valid, fill it in:
		if(this.getDoc() && this.getDoc().location
			&& (this.getDoc().location.protocol == 'http:' || this.getDoc().location.protocol == 'https:')) {
			// Set the textbox to the current url
			var textbox = document.getElementById('addExceptionTextbox');
			textbox.value = this.getDoc().location.protocol + '//' + this.getDoc().location.host;
			// Activate the button:
			document.getElementById('addExceptionButton').setAttribute('disabled', 'false');
		}
		
		// Assign the treeview:
		this.exceptionsTree.view = this.exceptionsTreeView;
		
		// The TreeSelection object:
		this.exceptionsTreeSelection = this.exceptionsTree.view.selection;
		
		// The TreeBox object:
		this.exceptionsTreeBox = this.exceptionsTree.treeBoxObject;
		
		// Sort is to be ascending if clicked first:
		this.exceptionsAscending = true;
	},

	exceptionsCheckURL: function(event) {
		var url = event.target.value;
		// Lenient check if url contains only protocol, domain and port number,
		// but no path information, GET variables, anchors or whitespace:
		var regExpURL = /^(http)(s)?(:\/\/)[^\s\/\?#]+$/;
		if(regExpURL.test(url)) {
			document.getElementById('addExceptionButton').setAttribute('disabled', 'false');
		} else {
			document.getElementById('addExceptionButton').setAttribute('disabled', 'true');
		}
	},
	
	exceptionsAdd: function(event) {
		var url = document.getElementById('addExceptionTextbox').value;
		
		// Check if the url is already in the list:
		if(this.inArray(this.exceptions, url))
			return;
		
		// Add the url to the list:
		this.exceptions.push(url);
		
		// Update the tree count and notify the tree:
		this.exceptionsTreeView.rowCount++;
		this.exceptionsTreeBox.rowCountChanged(this.exceptionsTreeView.rowCount, +1);
		this.exceptionsTreeBox.invalidate();

		// Update the preferences:
		this.setExceptions(this.exceptions);

		// Enable the "removeAllButton":
		document.getElementById('removeAllButton').setAttribute('disabled', 'false');
	},
	
	exceptionsSort: function(event) {
		// Sort the exception list:
		this.exceptions.sort();
		if(this.exceptionsAscending)
			this.exceptionsAscending = false;
		else {
			this.exceptions.reverse();
			this.exceptionsAscending = true;
		}
		
		// Notify the tree:
		this.exceptionsTreeBox.invalidate();
		
		// Clear out selections
		this.exceptionsTreeSelection.select(-1);
		
		// Disable "remove" button:
		document.getElementById('removeSelectedButton').setAttribute("disabled", "true");
	},
	
	exceptionsSelected: function(event) {
		if(this.exceptionsTreeSelection.count > 0) {
			document.getElementById('removeSelectedButton').setAttribute('disabled', 'false');
		}
	},
	
	exceptionsRemoveSelected: function(event) {
		// Start of update batch:
		this.exceptionsTreeBox.beginUpdateBatch();
			
		// Helper object to store a range:
		function Range(start, end) {
			this.start = start.value;
			this.end = end.value;
		}
		
		// List of ranges:
		var ranges = new Array();
		
		// Get the number of ranges:
		var numRanges = this.exceptionsTreeSelection.getRangeCount();
		
		// Helper vars to store the range end points:
		var start = new Object();
		var end = new Object();
		
		// We store the list of ranges first, as calling
		// this.exceptionsTreeBox.rowCountChanged()
		// seems to invalidate the current selection
		
		for(var i=0; i < numRanges; i++) {
			// Get the current range end points:
			this.exceptionsTreeSelection.getRangeAt(i,start,end);
			// Store them as a Range object in the ranges list:
			ranges[i] = new Range(start, end);
		}
		
		for(var i=0; i < numRanges; i++) {
			// Go through the stored ranges:
			for(var j = ranges[i].start; j <= ranges[i].end; j++) {
				// Set the selected exceptions to null:
				this.exceptions[j] = null;
			}
			
			// Calculate the new tree count:
			var count = ranges[i].end - ranges[i].start + 1;
			
			// Update the tree count and notify the tree:
			this.exceptionsTreeView.rowCount -= count;
			this.exceptionsTreeBox.rowCountChanged(ranges[i].start, -count);
		}

		// Collapse list by removing all the null entries
		for (var i=0; i < this.exceptions.length; i++) {
			if (!this.exceptions[i]) {
				var j = i;
				while (j < this.exceptions.length && !this.exceptions[j])
					j++;
				this.exceptions.splice(i, j-i);
			}
		}
		
		// Clear out selections
		this.exceptionsTreeSelection.select(-1); 
		
		// End of update batch:
		this.exceptionsTreeBox.endUpdateBatch();
		
		// Disable buttons:
		if(this.exceptions.length == 0) {
			document.getElementById('removeAllButton').setAttribute("disabled","true");
		}
		document.getElementById('removeSelectedButton').setAttribute("disabled", "true");
		
		// Update the preferences:
		this.setExceptions(this.exceptions);
	},
	
	exceptionsRemoveAll: function() {
		// The number of currently stored exceptions:
		var count = this.exceptions.length;
	
		// Empty the list:
		this.exceptions = new Array();

		// Clear out selections
		this.exceptionsTreeSelection.select(-1); 

		// Update the tree view and notify the tree
		this.exceptionsTreeView.rowCount = 0;
		// On deletion, notify from which index and how many rows have been deleted:
		this.exceptionsTreeBox.rowCountChanged(0, -count);
		this.exceptionsTreeBox.invalidate();

		// Disable buttons
		document.getElementById('removeSelectedButton').setAttribute("disabled", "true")
		document.getElementById('removeAllButton').setAttribute("disabled","true");

		// Update the preferences:
		this.setExceptions(this.exceptions);
	},

	exceptionsHandleKeyPress: function(event) {
		if(event.keyCode == 46) {
			this.exceptionsRemoveSelected();
		} else if(event.ctrlKey && event.which == 97) {
			if(this.exceptionsTree && this.exceptionsTreeSelection) {
				try {
					// Select all rows:
					this.exceptionsTreeSelection.selectAll();	
				} catch(e) {
					//alert(e);
				}
			}
		}
	},
	
	exceptionsDestroy: function() {
	},
	
	destroy: function()  {
		// Remove the content area context menu listener:
		if(document.getElementById('contentAreaContextMenu')) {
			document.getElementById('contentAreaContextMenu').removeEventListener(
				'popupshowing',
				this.contentAreaContextMenuEventListener,
				false
			);
		}
		
		// Remove the listener from the browser object:
		try {
			this.getBrowser().removeProgressListener(this.progressListener);
		} catch(e) {}
		
		// Remove the preferences Observer:
		this.getSecureLoginPrefs().removeObserver('', this);
		
		// Re-enable the prefilling of login forms if setting has been true:
		try {
			if(this.getVersionComparator().compare(this.getAppInfo().version, '2.*') < 0) {
				// Firefox version 1.5 - 2.0.0.*:
				if(this.autofillForms) {
					this.getPrefManager().getBranch('').setBoolPref('signon.prefillForms', true);
				}
			} else {
				// Firefox version 3 and prereleases:
				if(this.autofillForms) {
					this.getPrefManager().getBranch('').setBoolPref('signon.autofillForms', true);
				}
			}
		} catch(e) {
			//alert(e);
		}
	}
	
}